from nmigen import *
from nmigen.hdl.rec import *
from nmigen.cli import main
#from nmigen.utils import log2_int


from functools import reduce

from skidbuffer import RegisteredSkidBuffer

class ParseStackLayout(Layout):
    def __init__(self, *, data_width, index_width, command_width):
        super().__init__([
            # INPUTS
            ("command_in",     unsigned(command_width)), # FROM SOURCE
            ("data_in",        unsigned(data_width)),    # FROM SOURCE
            ("index_in",       unsigned(index_width)),   # FROM SOURCE
            ("valid_in",       1),                       # FROM SOURCE
            ("ready_in",       1),                       # FROM DEST

            # OUTPUTS
            ("data_out",       unsigned(data_width)),    # TO DEST
            ("index_out",      unsigned(index_width)),   # TO DEST
            ("valid_out",      1),                       # TO DEST
            ("internal_fault", 1),                       # TO DEST
            ("ready_out",      1),                       # TO SOURCE

        ])

class ParseStackBus(Record):
    def __init__(self, *, data_width, index_width, command_width):
        super().__init__(ParseStackLayout(data_width=data_width, index_width=index_width, command_width=command_width))

class ParseStack(Elaboratable):
    PUSH = 0
    POP  = 1
    MULTIPOP = 2
    INTROSPECT = 3
    READ_SP    = 4
    READ_TOS   = 5
    def __init__(self, width, depth):
        # Parameters
        self.width = width
        self.depth = depth

        # Control inputs
        self.command_port = Signal(3)
        # The commands are as follows:
        # 0: push
        # 1: pop
        # 2: multipop (number to pop in aux port)
        # 3: introspect (index in aux port)
        # 4: read stack pointer
        # 5: read top of stack

        # pop outputs the value of the popped item on the output data port

        # multipop outputs the value *at the top of the stack* after all the popping is done
        # on the output data port

        # read stack pointer outputs the stack pointer on the aux port
        dummy_index = Signal(range(depth))

        self.bus = ParseStackBus(data_width = width, index_width=len(dummy_index), command_width=3)

        #self.strobe       = Signal(1) # replace this with valid_in

        # Data inputs
        #self.in_data_port      = Signal(width) # data_in
        #self.in_aux_port       = Signal(range(depth)) #Index_in
        #self.in_data_valid     = Signal(1) # merge this and the one below, replace with valid_in
        self.in_aux_valid      = Signal(1) #


        # Control outputs
        self.internal_fault    = Signal(1) # keep

        # Data outputs
        self.out_data_port       = Signal(width) #data_out
        self.out_aux_port        = Signal(range(depth)) # index_out
        self.out_data_valid      = Signal(1) # merge these as valid_out
        self.out_aux_valid       = Signal(1)#


        self.mem = Memory(width=self.width, depth=self.depth)


    def elaborate(self, platform):
        m = Module()

        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()
        stack_pointer = Signal(range(self.depth), reset=0)


        # To represent the state of a k-item stack, we need an additional bit of state to
        # disambiguate the case where there are zero items on the stack and the case where
        # there is a single item pushed on the stack.

        # bottom_of_stack_valid = 0 implies zero items on the stack (which means s_p must be 0), and
        # popping generates an error

        # bottom_of_stack_valid = 1 implies that there is at least one item on the stack, and s_p can take
        # any value.

        # 0 items on stack (bottom_of_stack_valid = 0, stack_pointer = 0)
        # PUSH
        # 1 item on stack  (bottom_of_stack_valid = 1, stack_pointer = 0)
        # PUSH
        # 2 items on stack (bottom_of_stack_valid = 1, stack_pointer = 1)
        # PUSH
        # 3 items on stack (bottom_of_stack_valid = 1, stack_pointer = 2)

        # and for popping, it proceeds in reverse.

        bottom_of_stack_valid = Signal(1)


        with m.FSM() as fsm:
            with m.State("EMPTY"):
                with m.If(self.bus.valid_in == 1):
                    with m.Switch(self.bus.command_in):
                        with m.Case(self.PUSH):
                            m.d.comb += wport.addr.eq(stack_pointer)
                            m.d.comb += wport.data.eq(self.bus.data_in)
                            m.d.comb += wport.en.eq(1)
                            m.next = "AT_LEAST_ONE_ITEM"
                        with m.Case(self.POP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.MULTIPOP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.INTROSPECT):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_TOS):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"
                        with m.Case(self.READ_SP):
                            m.d.comb += self.internal_fault.eq(1)
                            m.next = "FAULTED"

            with m.State("AT_LEAST_ONE_ITEM"):
                with m.If(self.bus.valid_in == 1):
                    with m.Switch(self.command_port):
                        with m.Case(self.PUSH):
                            m.d.sync += stack_pointer.eq(stack_pointer + 1)
                            m.d.comb += wport.addr.eq(stack_pointer + 1)
                            m.d.comb += wport.data.eq(self.bus.data_in)
                            m.d.comb += wport.en.eq(1)
                            m.next = "AT_LEAST_ONE_ITEM"
                        with m.Case(self.POP):
                            m.d.comb += rport.addr.eq(stack_pointer)
                            m.d.comb += self.out_data_port.eq(rport.data)
                            m.d.comb += self.out_data_valid.eq(1)

                            with m.If(stack_pointer == 0):
                                m.next = "EMPTY"
                            with m.Else():
                                m.next = "AT_LEAST_ONE_ITEM"
                                m.d.sync += stack_pointer.eq(stack_pointer - 1)
                        with m.Case(self.MULTIPOP):
                            with m.If(self.in_aux_valid == 1):
                                with m.If((self.bus.index_in - 1 ) > stack_pointer):
                                    m.d.comb += self.internal_fault.eq(1)
                                    m.next = "FAULTED"
                                with m.Elif(self.bus.index_in == stack_pointer):
                                    m.d.sync += stack_pointer.eq(0)
                                    m.next = "EMPTY"
                                with m.Elif(self.bus.index_in < stack_pointer):
                                    m.d.sync += stack_pointer.eq(stack_pointer - self.bus.index_in)
                                    m.next = "AT_LEAST_ONE_ITEM"

                        with m.Case(self.INTROSPECT):
                            with m.If(self.in_aux_valid):
                                with m.If(self.bus.index_in > stack_pointer):
                                    m.d.comb += self.internal_fault.eq(1)
                                    m.next = "FAULTED"
                                with m.Else():
                                    m.d.comb += rport.addr.eq(self.bus.index_in)
                                    m.d.comb += self.out_data_port.eq(rport.data)
                                    m.d.comb += self.out_data_valid.eq(1)
                                    m.next = "AT_LEAST_ONE_ITEM"

                        with m.Case(self.READ_TOS):
                            m.d.comb += self.out_data_valid.eq(1)
                            m.d.comb += rport.addr.eq(stack_pointer)
                            m.d.comb += self.out_data_port.eq(rport.data)
                            m.d.comb += self.out_data_valid.eq(1)
                            m.next = "AT_LEAST_ONE_ITEM"

                        with m.Case(self.READ_SP):
                            m.d.comb += self.out_aux_port.eq(stack_pointer)
                            m.d.comb += self.out_aux_valid.eq(1)
                            m.next = "AT_LEAST_ONE_ITEM"

            with m.State("FAULTED"):
                m.d.comb += self.internal_fault.eq(1)
                m.next = "FAULTED"





 




        return m


class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.stack = ParseStack(8,8)
        counter = Signal(8)
        m.d.sync += counter.eq(counter+1)

        with m.If(counter==1):
            m.d.comb += m.submodules.stack.bus.valid_in.eq(1)
            m.d.comb += m.submodules.stack.bus.command_in.eq(ParseStack.PUSH)
            m.d.comb += m.submodules.stack.bus.data_in.eq(42)


        # with m.If(counter==2):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.PUSH)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)
        #     m.d.comb += m.submodules.stack.in_data_valid.eq(1)
        #     m.d.comb += m.submodules.stack.in_data_port.eq(43)

        # with m.If(counter==3):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.PUSH)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)
        #     m.d.comb += m.submodules.stack.in_data_valid.eq(1)
        #     m.d.comb += m.submodules.stack.in_data_port.eq(44)

        # with m.If(counter==4):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.POP)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)

        # with m.If(counter==5):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.POP)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)


        # with m.If(counter==7):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.POP)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)
        # with m.If(counter==9):
        #     m.d.comb += m.submodules.stack.command_port.eq(ParseStack.POP)
        #     m.d.comb += m.submodules.stack.strobe.eq(1)
        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
